version = "0.12.2"
project_url = "https://pypi.python.org/pypi/sftp-cloudfs/"
config_file = "/etc/sftpcloudfs.conf"

# keystone defaults
default_ks_tenant_separator = '.'
default_ks_service_type = 'object-store'
default_ks_endpoint_type = 'publicURL'
